/*
* Copyright (c) 2017 SecureRF Corporation, All rights reserved.
* -------------------------------------------------------------------------
* This computer program is proprietary information of SecureRF Corporation.
* The receipt or possession of this program does not convey any rights to
* reproduce or disclose its contents, or to manufacture, use, or sell
* anything that it may describe, in whole or in part, except for the purposes of 
* the post-quantum algorithm public review and evaluation in connection with 
* the National Institute of Standards and Technology’s development of standards 
* for post-quantum cryptography.  Any reproduction of this program for any other 
* purpose without the express written consent of SecureRF is a violation of the 
* copyright laws and may subject you to criminal prosecution.
* -------------------------------------------------------------------------
*/

#include "api.h"
#include <string.h>
#include <stddef.h>
#include <ae_crypto.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include "crypto_sign.h"

int crypto_sign_keypair(
	unsigned char *pk,
	unsigned char *sk
)
{

	unsigned char privkey[2*(WALNUT_VALUE_l + 2)];
	GF_T pubkey[WALNUT_PUBKEY_ELEMENTS];
	unsigned char * privkey2;

	int success = 1;

	success &= key_generation(privkey, pubkey);

	privkey2 = privkey + GET_NUM_BRAID_BYTES_PACKED(privkey);

	memcpy(pk, pubkey, WALNUT_PUBKEY_LENGTH);
	memcpy(sk, privkey, GET_NUM_BRAID_BYTES_PACKED(privkey) + GET_NUM_BRAID_BYTES_PACKED(privkey2));

#ifdef KAT_INTERMEDIATES
	unpack_and_print_matrix(pk, "Unpacked Public Key");
	printBuffer(pk, WALNUT_PUBKEY_LENGTH/sizeof(GF_T), "Packed Public Key");
#endif

	return !success;
}


int crypto_sign(
	unsigned char *sm, unsigned long long *smlen,
	const unsigned char *m, unsigned long long mlen,
	const unsigned char *sk
)
{
	int success = 1;

	unsigned char signature[CRYPTO_BYTES + mlen];

	success = signature_generation(signature, smlen, m, mlen, sk);

	memcpy(sm, signature, *smlen);

	return !success;
}

int crypto_sign_open(
	unsigned char *m, unsigned long long *mlen,
	const unsigned char *sm, unsigned long long smlen,
	const unsigned char *pk
)
{
	int success = 1;

	unsigned char *message = NULL;

	success &= signature_verification((unsigned char *)pk, (unsigned char *)sm, &smlen, &message, mlen);

	memcpy(m, message, *mlen);

	free(message);

	return !success;
}
